dbboost='''PRAGMA cache_size = 20000;'''
dbcreate='''
CREATE TABLE project (
    "name" TEXT NOT NULL,
    "description" TEXT,
    "kind" TEXT,
    "tmp" TEXT,
    "creation" TEXT,
    "last" TEXT,
    "genome" TEXT,
    "phenome" TEXT,
    "pangenome" INTEGER   DEFAULT (0),
    "kegg" REAL
);
CREATE TABLE organism (
    "org_id" TEXT NOT NULL,
    "name" TEXT,
    "description" TEXT,
    "file" TEXT,
    "mutant" INTEGER,
    "reference" TEXT,
    "genome" TEXT   DEFAULT ('none'),
    "phenome" TEXT   DEFAULT ('none'),
    "mkind" TEXT,
	"color" TEXT
);
CREATE TABLE "protein" (
    "prot_id" TEXT NOT NULL,
    "org_id" TEXT NOT NULL,
    "description" TEXT,
    "sequence" TEXT
);
CREATE TABLE "mapko" (
    "prot_id" TEXT,
    "ko_id" TEXT,
    "indirect" INTEGER DEFAULT (0)
);
CREATE TABLE "ortholog" (
    "group_id" TEXT,
    "prot_id" TEXT
);
CREATE TABLE "ko" (
    "ko_id" TEXT NOT NULL,
    "name" TEXT,
    "description" TEXT,
    "analyzed" INTEGER   DEFAULT (0)
);
CREATE TABLE "reaction" (
    "re_id" TEXT NOT NULL,
    "name" TEXT,
    "description" TEXT,
    "enzyme" TEXT
);
CREATE TABLE "rpair" (
    "rp_id" TEXT NOT NULL,
    "co1" TEXT,
    "co2" TEXT,
    "kind" TEXT 
);
CREATE TABLE "compound" (
    "co_id" TEXT NOT NULL,
    "name" TEXT,
    "description" TEXT
);
CREATE TABLE "pathway" (
    "path_id" TEXT NOT NULL,
    "name" TEXT,
    "description" TEXT,
    "html" TEXT
);
CREATE TABLE "ko_react" (
    "ko_id" TEXT NOT NULL,
    "re_id" TEXT NOT NULL
);
CREATE TABLE "react_comp" (
    "re_id" TEXT NOT NULL,
    "co_id" TEXT NOT NULL
);
CREATE TABLE "react_path" (
    "re_id" TEXT NOT NULL,
    "path_id" TEXT NOT NULL
);
CREATE TABLE "comp_path" (
    "co_id" TEXT NOT NULL,
    "path_id" TEXT NOT NULL
);
CREATE TABLE "rpair_react" (
    "rp_id" TEXT NOT NULL,
    "re_id" TEXT NOT NULL
);
CREATE TABLE "pathmap" (
    "path_id" TEXT NOT NULL,
    "png" BLOB,
    "html" TEXT
);
CREATE TABLE biolog (
    "plate_id" TEXT NOT NULL,
    "well_id" TEXT NOT NULL,
    "concentration" INTEGER,
    "zero_well_id" TEXT,
    "chemical" TEXT,
    "cas_id" TEXT,
    "co_id" TEXT,
    "moa" TEXT
, "category" INTEGER
);
CREATE TABLE biolog_exp (
    "plate_id" TEXT NOT NULL,
    "well_id" TEXT NOT NULL,
    "org_id" TEXT NOT NULL,
    "replica" INTEGER NOT NULL,
    "activity" INTEGER,
    "zero" INTEGER,
    "min" REAL,
    "max" REAL,
    "height" REAL,
    "plateau" REAL,
    "slope" REAL,
    "lag" REAL,
    "area" REAL,
    "v" REAL,
    "y0" REAL,
    "model" TEXT,
    "source" TEXT
);
CREATE TABLE biolog_exp_det (
    "plate_id" TEXT NOT NULL,
    "well_id" TEXT NOT NULL,
    "org_id" TEXT NOT NULL,
    "replica" INTEGER NOT NULL,
    "times" TEXT,
    "signals" TEXT
);
CREATE TABLE biolog_purged_exp (
    "plate_id" TEXT NOT NULL,
    "well_id" TEXT NOT NULL,
    "org_id" TEXT NOT NULL,
    "replica" INTEGER NOT NULL,
    "activity" INTEGER,
    "zero" INTEGER,
    "min" REAL,
    "max" REAL,
    "height" REAL,
    "plateau" REAL,
    "slope" REAL,
    "lag" REAL,
    "area" REAL,
    "v" REAL,
    "y0" REAL,
    "model" TEXT,
    "source" TEXT
);
CREATE TABLE biolog_purged_exp_det (
    "plate_id" TEXT NOT NULL,
    "well_id" TEXT NOT NULL,
    "org_id" TEXT NOT NULL,
    "replica" INTEGER NOT NULL,
    "times" TEXT,
    "signals" TEXT
);
CREATE UNIQUE INDEX project_id ON project(name ASC);
CREATE UNIQUE INDEX "organism_id" on organism (org_id ASC);
CREATE UNIQUE INDEX "protein_id" on protein (prot_id ASC, org_id ASC);
CREATE UNIQUE INDEX "mapko_id" on mapko (prot_id ASC, ko_id ASC);
CREATE UNIQUE INDEX "ortholog_id" on ortholog (group_id ASC, prot_id ASC);
CREATE UNIQUE INDEX "ko_id" on ko (ko_id ASC);
CREATE UNIQUE INDEX "reaction_id" on reaction (re_id ASC);
CREATE UNIQUE INDEX "rpair_id" on rpair (rp_id ASC);
CREATE UNIQUE INDEX "compound_id" on compound (co_id ASC);
CREATE UNIQUE INDEX "pathway_id" on pathway (path_id ASC);
CREATE UNIQUE INDEX "koreact_id" on ko_react (ko_id ASC, re_id ASC);
CREATE UNIQUE INDEX "reactcomp_id" on react_comp (re_id ASC, co_id ASC);
CREATE UNIQUE INDEX "reactpath_id" on react_path (re_id ASC, path_id ASC);
CREATE UNIQUE INDEX "comppath_id" on comp_path (co_id ASC, path_id ASC);
CREATE UNIQUE INDEX "rpairreact_id" on rpair_react (rp_id ASC, re_id ASC);
CREATE UNIQUE INDEX "pathmap_id" on pathmap (path_id ASC);
CREATE UNIQUE INDEX "biolog_id" on biolog (plate_id ASC, well_id ASC);
CREATE UNIQUE INDEX "biologexp_id" on biolog_exp (plate_id ASC, well_id ASC, org_id ASC, replica ASC);
CREATE UNIQUE INDEX "biologexpdet_id" on biolog_exp_det (plate_id ASC, well_id ASC, org_id ASC, replica ASC);
CREATE UNIQUE INDEX "biologpurgedexp_id" on biolog_purged_exp (plate_id ASC, well_id ASC, org_id ASC, replica ASC);
CREATE UNIQUE INDEX "biologpurgedexpdet_id" on biolog_purged_exp_det (plate_id ASC, well_id ASC, org_id ASC, replica ASC);
'''



#!/usr/bin/env python
"""
Actions Parallel

DuctApe Library

All the actions required for the analysis (parallel utilitites)
"""
from ductape.common.commonmultiprocess import CommonMultiProcess
from ductape.storage.SQLite.database import Kegg
import logging
import sys
if sys.version_info[0] < 3:
    import Queue as queue
else:
    import queue

__author__ = "Marco Galardini"

################################################################################
# Log setup

logger = logging.getLogger('ductape.actionsparallel')

################################################################################
# Classes

class RunPath(object):
    def __init__(self, project, path_id):
        self._project = project
        self.path_id = path_id
    
    def __call__(self):
        db = Kegg(self._project)
        allr = [r for r in db.getMappedRPairsReact(self.path_id)]
        ecore, edisp, eacc, euni = db.getExclusiveRPairsReact(self.path_id)
        dpangenome = {'all':allr,
                        'core':ecore, 'dispensable':edisp,
                          'accessory':eacc, 'unique':euni}

        return self.path_id, dpangenome

class PathPanGenomer(CommonMultiProcess):
    '''
    Class PathPangenomer
    '''
    _statusDesc = {0:'Not started',
               1:'Analyzing pangenomic pathways'}
    
    _substatuses = [1]
    
    def __init__(self,project, paths,
                 ncpus=1,queue=queue.Queue()):
        CommonMultiProcess.__init__(self,ncpus,queue)
        
        # DB name
        self._project = project
        
        # To-be-analyzed pathways
        self.paths = paths
        
        # path_id --> dpangenome
        self.result = {}
    
    def analyzePaths(self):
        self._maxsubstatus = len(self.paths)
        
        self.initiateParallel()
        for path in self.paths:
            # Multi process
            obj = RunPath(self._project, path)
            self._paralleltasks.put(obj)
                
        # Poison pill to stop the workers
        self.addPoison()
        
        while True:
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return
                 
            while not self._parallelresults.empty():
                if self.killed:
                    logger.debug('Exiting for a kill signal')
                    return
                
                path_id, dpangenome = self._parallelresults.get()
                self.result[path_id] = dpangenome
                
                self._substatus += 1
                self.updateStatus(sub=True)
                    
            if self.isTerminated():
                break
            
            self.sleeper.sleep(0.1)
            
        while not self._parallelresults.empty():
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return
            
            path_id, dpangenome = self._parallelresults.get()
            self.result[path_id] = dpangenome
            
            self._substatus += 1
            self.updateStatus(sub=True)
        
        self.killParallel()
        
        return True
    
    def run(self):
        self.updateStatus()
        if not self.analyzePaths():
            self.sendFailure('Could not analyze pathways!')
            return
        self.resetSubStatus()



#!/usr/bin/env python
"""
TerminalProgress

Common library

Notify in a shell the progress of something

Borrowed from Calibre linux installer
"""

import re
import sys

class TerminalController:
    BOL = ''             #: Move the cursor to the beginning of the line
    UP = ''              #: Move the cursor up one line
    DOWN = ''            #: Move the cursor down one line
    LEFT = ''            #: Move the cursor left one char
    RIGHT = ''           #: Move the cursor right one char

    # Deletion:
    CLEAR_SCREEN = ''    #: Clear the screen and move to home position
    CLEAR_EOL = ''       #: Clear to the end of the line.
    CLEAR_BOL = ''       #: Clear to the beginning of the line.
    CLEAR_EOS = ''       #: Clear to the end of the screen

    # Output modes:
    BOLD = ''            #: Turn on bold mode
    BLINK = ''           #: Turn on blink mode
    DIM = ''             #: Turn on half-bright mode
    REVERSE = ''         #: Turn on reverse-video mode
    NORMAL = ''          #: Turn off all modes

    # Cursor display:
    HIDE_CURSOR = ''     #: Make the cursor invisible
    SHOW_CURSOR = ''     #: Make the cursor visible

    # Terminal size:
    COLS = None          #: Width of the terminal (None for unknown)
    LINES = None         #: Height of the terminal (None for unknown)

    # Foreground colors:
    BLACK = BLUE = GREEN = CYAN = RED = MAGENTA = YELLOW = WHITE = ''

    # Background colors:
    BG_BLACK = BG_BLUE = BG_GREEN = BG_CYAN = ''
    BG_RED = BG_MAGENTA = BG_YELLOW = BG_WHITE = ''

    _STRING_CAPABILITIES = """
    BOL=cr UP=cuu1 DOWN=cud1 LEFT=cub1 RIGHT=cuf1
    CLEAR_SCREEN=clear CLEAR_EOL=el CLEAR_BOL=el1 CLEAR_EOS=ed BOLD=bold
    BLINK=blink DIM=dim REVERSE=rev UNDERLINE=smul NORMAL=sgr0
    HIDE_CURSOR=cinvis SHOW_CURSOR=cnorm""".split()
    _COLORS = """BLACK BLUE GREEN CYAN RED MAGENTA YELLOW WHITE""".split()
    _ANSICOLORS = "BLACK RED GREEN YELLOW BLUE MAGENTA CYAN WHITE".split()

    def __init__(self, term_stream=sys.stdout):
        # Curses isn't available on all platforms
        try: import curses
        except: return

        # If the stream isn't a tty, then assume it has no capabilities.
        if not hasattr(term_stream, 'isatty') or not term_stream.isatty(): return

        # Check the terminal type.  If we fail, then assume that the
        # terminal has no capabilities.
        try: curses.setupterm()
        except: return

        # Look up numeric capabilities.
        self.COLS = curses.tigetnum('cols')
        self.LINES = curses.tigetnum('lines')

        # Look up string capabilities.
        for capability in self._STRING_CAPABILITIES:
            (attrib, cap_name) = capability.split('=')
            setattr(self, attrib, self._tigetstr(cap_name).encode() or b'')

        # Colors
        set_fg = self._tigetstr('setf').encode()
        if set_fg:
            for i,color in zip(list(range(len(self._COLORS))), self._COLORS):
                setattr(self, color, curses.tparm(set_fg, i) or b'')
        set_fg_ansi = self._tigetstr('setaf').encode()
        if set_fg_ansi:
            for i,color in zip(list(range(len(self._ANSICOLORS))), self._ANSICOLORS):
                setattr(self, color, curses.tparm(set_fg_ansi, i) or b'')
        set_bg = self._tigetstr('setb').encode()
        if set_bg:
            for i,color in zip(list(range(len(self._COLORS))), self._COLORS):
                setattr(self, 'BG_'+color, curses.tparm(set_bg, i) or b'')
        set_bg_ansi = self._tigetstr('setab').encode()
        if set_bg_ansi:
            for i,color in zip(list(range(len(self._ANSICOLORS))), self._ANSICOLORS):
                setattr(self, 'BG_'+color, curses.tparm(set_bg_ansi, i) or b'')

    def _tigetstr(self, cap_name):
        # String capabilities can include "delays" of the form "$<2>".
        # For any modern terminal, we should be able to just ignore
        # these, so strip them out.
        import curses
        cap = curses.tigetstr(cap_name) or b''
        return re.sub(r'\$<\d+>[/*]?', '', cap.decode('utf-8'))

    def render(self, template):
        return re.sub(r'\$\$|\${\w+}', self._render_sub, template)

    def _render_sub(self, match):
        s = match.group()
        if s == '$$': return s
        else:
            s = getattr(self, s[2:-1])
            if isinstance(s, bytes):
                s = s.decode('utf-8')
            return s

class ProgressBar:
    BAR = '%3d%% ${GREEN}[${BOLD}%s%s${NORMAL}${GREEN}]${NORMAL}\n'
    HEADER = '${BOLD}${BLUE}%s${NORMAL}\n\n'

    def __init__(self, term, header):
        self.term = term
        if not (self.term.CLEAR_EOL and self.term.UP and self.term.BOL):
            self.use_it = False
        else:
            self.use_it = True
        self.width = self.term.COLS or 75
        self.bar = term.render(self.BAR)
        self.header = self.term.render(self.HEADER % header.center(self.width))
        self.cleared = 1 #: true if we haven't drawn the bar yet.

    def update(self, percent, message=''):
        if not self.use_it:
            return
        if isinstance(message, str):
            message = message.encode('utf-8', 'ignore')
        if self.cleared:
            sys.stdout.write(self.header)
            self.cleared = 0
        n = int((self.width-10)*percent)
        msg = message.center(self.width)
        s = self.term.BOL + self.term.UP + self.term.CLEAR_EOL
        s = s.decode('utf-8')
        sys.stdout.write(
        s +
        (self.bar % (100*percent, '='*n, '-'*(self.width-10-n))) +
        self.term.CLEAR_EOL.decode('utf-8') + msg.decode('utf-8'))
        sys.stdout.flush()

    def clear(self):
        if not self.use_it:
            return
        if not self.cleared:
            sys.stdout.write(self.term.BOL + self.term.CLEAR_EOL +
            self.term.UP + self.term.CLEAR_EOL +
            self.term.UP + self.term.CLEAR_EOL)
            self.cleared = 1



#!/usr/bin/env python
"""
Kegg

Kegg Library

Kegg data fetching
"""
import sys
if sys.version_info[0] < 3:
    import Queue as queue
    from urllib2 import quote
    from urllib2 import urlopen
else:
    from urllib.request import urlopen
    from urllib.parse import quote
    import queue
from ductape.common.commonthread import CommonThread
from ductape.common.utils import get_span
from ductape.common.utils import isOnline
from ductape.kegg.web import kheader
from matplotlib import colors
import logging
import os
import shutil
import threading
import time
import random

__author__ = "Marco Galardini"

################################################################################
# Log setup

logger = logging.getLogger('ductape.kegg')

################################################################################
# Constants

avoidedPaths = set(['%s%s%s' % (prefix1, prefix2, value)
                    for prefix1 in ['', 'path:']
                    for prefix2 in ['rn', 'ko', 'map']
                    for value in ['01100', '01110', '01120',
                                  '01200', '01210', '01212',
                                  '01230', '01220']
                   ])

################################################################################
# Classes

class MapParser(object):
    '''
    Takes an HTML of a Kegg map and returns a list of ready-to-use html lines
    '''
    def __init__(self,content):
        self.html = content
        self.map = []
        self.parse()
        
    def parse(self):
        b = False
        for l in self.html.split('\n'):
            l = l.lstrip().rstrip().strip()
            if '<map' in l[:4]:
                b = True
                self.map.append(l)
            elif '<area' in l[:6] and b:
                self.map.append(l)
            elif '</map' in l[:5] and b:
                self.map.append(l)
                break
            
        return self.map

class KeggAPI(object):
    '''
    Class KeggAPI
    Connects to KEGG API and performs various tasks
    Fail-safe: if a request fails it tries again and again
    All the results are stored in the attribute result, as well as the inputs
    
    Docs:
    http://www.kegg.jp/kegg/docs/keggapi.html
    http://www.kegg.jp/kegg/rest/weblink.html
    '''
    def __init__(self, keeptrying=False):
        self.baseurl = 'http://www.kegg.jp/'
        self._apiurl = 'http://rest.kegg.jp/'
        self._maplink = 'http://www.kegg.jp/kegg-bin/show_pathway?'
        
        self.failed = False
        
        self.keeptrying = keeptrying
        
        self.clean()
    
    def clean(self):
        self.input = None
        self.result = None
        self.failed = False

    def getEntryTag(self, entry, tag):
        '''
        Get the tag content inside a kegg entry (flat file)
        '''
        b = False
        res = ''
        for line in entry.split('\n'):
            if line.startswith(tag):
                res += line.rstrip().lstrip(tag).lstrip()
                b = True
            elif b and line[0] == ' ':
                res += ' ' + line.rstrip().lstrip()
            elif b and line[0] != ' ':
                b = False
                return res
        
        if res == '':
            return None
        return res.lstrip()
    
    def getLinkTag(self, entry, tag):
        '''
        Get the tag content inside a kegg entry (flat file)
        This variant function extract links from an entry
        '''
        b = False
        for line in entry.split('\n'):
            if line.startswith(tag):
                yield line.rstrip().lstrip(tag).lstrip().split()[-1]
                b = True
            elif b and line[0] == ' ':
                yield line.rstrip().lstrip().split()[-1]
            elif b and line[0] != ' ':
                b = False
        
    def parseLinks(self, links):
        '''
        Parse the results of 
        '''
        d = {}
        for line in links.split('\n'):
            if line == '' or '\t' not in line:continue
            
            k, v = line.split('\t')
            
            if k not in d:
                d[k] = []
            d[k].append(v)
        
        if len(d) == 0:
            return None
        return d

    def getRelease(self, dbver):
        '''
        Takes a string like "Release 64.0+/11-13, Nov 12"
        and returns the release number, as a float (useful for comparison)
        if it fails returns None
        '''
        try:
            s = dbver.split(' ')
            release = s[1]

            while True:
                try:
                    release = float(release)
                    return release
                except:
                    release = release[:-1]
        except:
            logger.debug('Could not parse KEGG database version (%s)'%dbver)
            return None

    def getDBVersion(self, retries=8):
        '''
        Get the KEGG DB version
        Returns a tuple: full version string , release number (None if unparsable)
        '''
        attempts = 0
        while True:
            try:
                self.input = None
                logger.debug('Looking for KEGG db version')
                url = self._apiurl + quote('info/kegg')
                data = urlopen(url, timeout=20).read().decode('utf-8').split('\n')
                
                line = data[1].split('             ')[1]
                self.result = (line, self.getRelease(line))

                return
            except Exception as e:
                attempts += 1
                logger.debug('info failed! Attempt %d'
                              %attempts)
                logger.debug('%s'%str(e))
                time.sleep((2 + random.random())*attempts)
                try:
                    logger.debug(url)
                except:pass
                if self.keeptrying:continue
                if attempts >= retries:
                    self.failed = True
                    logger.warning('info failed!')
                    return
    
    def getTitle(self, entries, otherTags=[], retries=8):
        '''
        Get the title of a specific KEGG object
        Default behaviour is to return NAME and DEFINITION tags
        Additional tags can be provided in otherTags
        '''
        attempts = 0
        while True:
            try:
                self.input = entries
                logger.debug('Looking for title for %d KEGG entries'%len(entries))
                url = ''
                for entry in entries:
                    url += '%s+'%entry
                    
                # Dummy entry to avoid a rare bug when all the provided entries
                if 'cpd:C00099' not in entries or 'C00099' not in entries:
                    url += 'cpd:C00099'
                #
                
                url = url.rstrip('+')
                url = self._apiurl + 'get/' + quote(url)
                data = urlopen(url, timeout=20).read().decode('utf-8')
                
                self.result = {}
                for lines in data.split('///'):
                    if len(lines) == 1:continue
                    try:
                        shortID = self.getEntryTag(lines,'ENTRY').split(' ')[0]
                        for longID in self.input:
                            if shortID in longID:
                                self.result[longID] = [
                                            self.getEntryTag(lines, 'NAME'),
                                            self.getEntryTag(lines, 'DEFINITION')
                                            ]

                                # TODO: a more general approach for this                                             
                                for tag in otherTags:
                                    value = self.getEntryTag(lines, tag)
                                    self.result[longID].append(value)
                    except:
                        continue
                    
                # Check that every input has a result
                for entry in entries:
                    if entry not in self.result:
                        self.result[entry] = ['','']
                
                return
            except Exception as e:
                attempts += 1
                logger.debug('get failed! Attempt %d'
                              %attempts)
                logger.debug('%s'%str(e))
                time.sleep((2 + random.random())*attempts)
                try:
                    logger.debug(url)
                except:pass
                if self.keeptrying:continue
                if attempts >= retries:
                    self.failed = True
                    logger.warning('get failed!')
                    return
          
    def getRPair(self, entries, retries=8):
        '''
        Similar to getTitle, but targeting specific features of RPair
        Specifically the two co_id involved and the kind of interaction
        '''
        attempts = 0
        while True:
            try:
                self.input = entries
                logger.debug('Looking for details on %d RPair entries'%len(entries))
                url = ''
                for entry in entries:
                    url += '%s+'%entry
                
                # Dummy entry to avoid a rare bug when all the provided entries
                if 'rp:RP00001' not in entries or 'RP00001' not in entries:
                    url += 'rp:RP00001'
                #
                    
                url = url.rstrip('+')
                url = self._apiurl + 'get/' + quote(url)
                data = urlopen(url, timeout=20).read().decode('utf-8')
                
                self.result = {}
                for lines in data.split('///'):
                    if len(lines) == 1:continue
                    try:
                        shortID = self.getEntryTag(lines,'ENTRY').split(' ')[0]
                        for longID in self.input:
                            if shortID in longID:
                                co1, co2 = self.getEntryTag(lines, 'NAME').split('_')
                                #
                                if not co1.startswith('cpd:'):
                                    co1 = 'cpd:' + co1
                                if not co2.startswith('cpd:'):
                                    co2 = 'cpd:' + co2
                                #
                                kind = self.getEntryTag(lines, 'TYPE')
                                self.result[longID] = [co1,co2,kind]
                    except:
                        continue
                    
                # Check that every input has a result
                for entry in entries:
                    if entry not in self.result:
                        self.result[entry] = ['','','']
                 
                return
            except Exception as e:
                attempts += 1
                logger.debug('get (rpair) failed! Attempt %d'
                              %attempts)
                logger.debug('%s'%str(e))
                time.sleep((2 + random.random())*attempts)
                try:
                    logger.debug(url)
                except:pass
                if self.keeptrying:continue
                if attempts >= retries:
                    self.failed = True
                    logger.warning('get (rpair) failed!')
                    return
    
    def getIDListFromDB(self, db='pathway', retries=8):
        '''
        Get all the IDs from a specific database
        
        Default: pathway
        '''
        attempts = 0
        while True:
            try:
                self.input = db
                logger.debug('Looking for KEGG IDs from db %s'%db)
                url = self._apiurl + 'list/%s/' % quote(db)
                
                data = urlopen(url, timeout=20).read().decode('utf-8')
                self.result = set([x.split('\t')[0] for x in data.split('\n')])
                try:
                    self.result.remove('')
                except:pass
                self.result = list(self.result)
                return
            except Exception as e:
                attempts += 1
                logger.debug('list (%s) failed! Attempt %d'
                              %(db,attempts))
                logger.debug('%s'%str(e))
                time.sleep((2 + random.random())*attempts)
                try:
                    logger.debug(url)
                except:pass
                if self.keeptrying:continue
                if attempts >= retries:
                    self.failed = True
                    logger.warning('list (%s) failed!'%db)
                    return
    
    def getReactions(self, ko_ids, retries=8):
        '''
        Get the reaction IDs for a given KO list
        '''
        attempts = 0
        while True:
            try:
                self.input = ko_ids
                logger.debug('Looking for KEGG reactions from %d KO IDs'%len(ko_ids))
                url = ''
                for ko_id in ko_ids:
                    url += '%s+'%ko_id
                url = url.rstrip('+')
                
                url = self._apiurl + 'link/reaction/' + quote(url)
                
                data = urlopen(url, timeout=20).read().decode('utf-8')
                self.result = self.parseLinks(data)
                return
            except Exception as e:
                attempts += 1
                logger.debug('link (reaction) failed! Attempt %d'
                              %attempts)
                logger.debug('%s'%str(e))
                time.sleep((2 + random.random())*attempts)
                try:
                    logger.debug(url)
                except:pass
                if self.keeptrying:continue
                if attempts >= retries:
                    self.failed = True
                    logger.warning('link (reaction) failed!')
                    return
                
    def getPathways(self, re_ids, retries=8):
        '''
        Get the pathway IDs for a given reaction list
        '''
        attempts = 0
        while True:
            try:
                self.input = re_ids
                logger.debug('Looking for KEGG pathways from %d RE IDs'%len(re_ids))
                url = ''
                for re_id in re_ids:
                    url += '%s+'%re_id
                url = url.rstrip('+')
                
                url = self._apiurl + 'link/pathway/' + quote(url)
                
                data = urlopen(url, timeout=20).read().decode('utf-8')
                self.result = self.parseLinks(data)
                return
            except Exception as e:
                attempts += 1
                logger.debug('link (pathway) failed! Attempt %d'
                              %attempts)
                logger.debug('%s'%str(e))
                time.sleep((2 + random.random())*attempts)
                try:
                    logger.debug(url)
                except:pass
                if self.keeptrying:continue
                if attempts >= retries:
                    self.failed = True
                    logger.warning('link (pathway) failed!')
                    return
    
    def getReactionsByComp(self, co_ids, retries=8):
        '''
        Get the reactions IDs for a given compound list
        '''
        attempts = 0
        while True:
            try:
                self.input = co_ids
                logger.debug('Looking for KEGG reactions from %d CO IDs'%len(co_ids))
                url = ''
                for co_id in co_ids:
                    url += '%s+'%co_id
                url = url.rstrip('+')
                
                url = self._apiurl + 'link/reaction/' + quote(url)
                
                data = urlopen(url, timeout=20).read().decode('utf-8')
                self.result = self.parseLinks(data)
                return
            except Exception as e:
                attempts += 1
                logger.debug('link (reaction) failed! Attempt %d'
                              %attempts)
                logger.debug('%s'%str(e))
                time.sleep((2 + random.random())*attempts)
                try:
                    logger.debug(url)
                except:pass
                if self.keeptrying:continue
                if attempts >= retries:
                    self.failed = True
                    logger.warning('link (reaction) failed!')
                    return
    
    def getReactionsFromPath(self, path_ids, retries=8):
        '''
        Get the reaction IDs for a given pathway list
        '''
        attempts = 0
        while True:
            try:
                self.input = path_ids
                logger.debug('Looking for KEGG reactions from %d PATH IDs'%len(path_ids))
                url = ''
                for path_id in path_ids:
                    url += '%s+'%path_id
                url = url.rstrip('+')
                
                url = self._apiurl + 'link/reaction/' + quote(url)
                
                data = urlopen(url, timeout=20).read().decode('utf-8')
                self.result = self.parseLinks(data)
                return
            except Exception as e:
                attempts += 1
                logger.debug('link (reaction) failed! Attempt %d'
                              %attempts)
                logger.debug('%s'%str(e))
                time.sleep((2 + random.random())*attempts)
                try:
                    logger.debug(url)
                except:pass
                if self.keeptrying:continue
                if attempts >= retries:
                    self.failed = True
                    logger.warning('link (reaction) failed!')
                    return
    
    def getRPairsFromReaction(self, entries, retries=8):
        '''
        Get the rpair IDs for a given reaction list
        '''
        attempts = 0
        while True:
            try:
                self.input = entries
                logger.debug('Looking for RClass for %d KEGG entries'%len(entries))
                url = ''
                for entry in entries:
                    url += '%s+'%entry
                    
                # Dummy entry to avoid a rare bug when all the provided entries
                if 'cpd:C00099' not in entries or 'C00099' not in entries:
                    url += 'cpd:C00099'
                #
                
                url = url.rstrip('+')
                url = self._apiurl + 'get/' + quote(url)
                data = urlopen(url, timeout=20).read().decode('utf-8')
                
                self.result = {}
                for lines in data.split('///'):
                    if len(lines) == 1:continue
                    try:
                        shortID = self.getEntryTag(lines,'ENTRY').split(' ')[0]
                        for longID in self.input:
                            if shortID in longID:
                                for rclass in self.getLinkTag(lines, 'RCLASS'):
                                    self.result[longID] = self.result.get(longID,
                                                                          set())
                                    self.result[longID].add(rclass)
                    except:
                        continue
                    
                return
            except Exception as e:
                attempts += 1
                logger.debug('link (rpair) failed! Attempt %d'
                              %attempts)
                logger.debug('%s'%str(e))
                time.sleep((2 + random.random())*attempts)
                try:
                    logger.debug(url)
                except:pass
                if self.keeptrying:continue
                if attempts >= retries:
                    self.failed = True
                    logger.warning('link (rpair) failed!')
                    return
                
    def getCompoundsFromReaction(self, re_ids, retries=8):
        '''
        Get the compound IDs for a given reaction list
        '''
        attempts = 0
        while True:
            try:
                self.input = re_ids
                logger.debug('Looking for KEGG compounds from %d RE IDs'%len(re_ids))
                url = ''
                for re_id in re_ids:
                    url += '%s+'%re_id
                url = url.rstrip('+')
                
                url = self._apiurl + 'link/compound/' + quote(url)
                
                data = urlopen(url, timeout=20).read().decode('utf-8')
                self.result = self.parseLinks(data)
                return
            except Exception as e:
                attempts += 1
                logger.debug('link (compound) failed! Attempt %d'
                              %attempts)
                logger.debug('%s'%str(e))
                time.sleep((2 + random.random())*attempts)
                try:
                    logger.debug(url)
                except:pass
                if self.keeptrying:continue
                if attempts >= retries:
                    self.failed = True
                    logger.warning('link (compound) failed!')
                    return
    
    def getCompoundsFromPath(self, path_ids, retries=8):
        '''
        Get the compound IDs for a given pathway list
        '''
        attempts = 0
        while True:
            try:
                self.input = path_ids
                logger.debug('Looking for KEGG compounds from %d PATH IDs'%len(path_ids))
                url = ''
                for path_id in path_ids:
                    url += '%s+'%path_id
                url = url.rstrip('+')
                
                url = self._apiurl + 'link/compound/' + quote(url)
                
                data = urlopen(url, timeout=20).read().decode('utf-8')
                self.result = self.parseLinks(data)
                return
            except Exception as e:
                attempts += 1
                logger.debug('link (compound) failed! Attempt %d'
                              %attempts)
                logger.debug('%s'%str(e))
                time.sleep((2 + random.random())*attempts)
                try:
                    logger.debug(url)
                except:pass
                if self.keeptrying:continue
                if attempts >= retries:
                    self.failed = True
                    logger.warning('link (compound) failed!')
                    return
    
    def getHTMLColoredPathway(self, path_id, obj_list, color_list,
                                    border_list=None, retries=8):
        '''
        Get the URL of the colored pathway and return its content
        If it fails, an exception is thrown
        '''
        attempts = 0
        while True:
            try:
                self.input = path_id
                
                # Fix color codes
                hblue = colors.cnames['blue'].replace('#', '%23')
                for i in range(len(color_list)):
                    if '#' in color_list[i]:
                        color_list[i] = color_list[i].replace('#', '%23')
                #
                
                logger.debug('Looking for KEGG colored map from %s'%path_id)
                url = path_id.lstrip('path:') + '/default%3dblue/'
                for i in range(len(obj_list)):
                    url += obj_list[i]
                    # do not waste precious characters
                    if color_list[i] == hblue:
                        if border_list is not None and border_list[i] is not None:
                            url += '%09,' + border_list[i] + '/'
                        else:
                            url += '/'
                    #
                    else:
                        if border_list is not None and border_list[i] is not None:
                            url += '%09' + color_list[i] + ',' + border_list[i] + '/'
                        else:
                            url += '%09' + color_list[i] + '/'
               
                # Cannot quote this url, no colored pathway can be obtained then
                #url = self._maplink + quote(url)
                url = self._maplink + url
                
                logger.debug(url)
                # check that URL length is below 2000
                if len(url) > 2000:
                    logger.warning('URL too long for pathway %s, will skip'%path_id)
                    self.result = ''
                    return

                sock=urlopen(url, timeout=60)
                self.result = sock.read().decode('utf-8')
                sock.close()
                return
            except Exception as e:
                attempts += 1
                logger.debug('show_pathway failed! Attempt %d'
                              %attempts)
                logger.debug('%s'%str(e))
                time.sleep((2 + random.random())*attempts)
                if self.keeptrying:continue
                if attempts >= retries:
                    self.failed = True
                    logger.warning('show_pathway failed!')
                    return

class KeggColor(object):
    '''
    Class KeggColor
    Holds the color information to be passed to MapsFetcher
    One object for each pathway
    '''
    def __init__(self, path, htmlmap= '', reactions={}, compounds={},
                 borders={}):
        self.path = path 
        self.htmlmap = htmlmap
        self.reactions = reactions
        self.compounds = compounds
        # Objects that need to have a coloured border
        self.borders = borders
    
    def setMap(self, htmlmap):
        self.htmlmap = htmlmap
    
    def setReactions(self, reactions):
        self.reactions = reactions
        
    def setCompounds(self, compounds):
        self.compounds = compounds
        
    def setBorders(self, borders):
        self.borders = borders
        
    def getAll(self):
        '''
        Returns a tuple --> objects, color
        '''
        objs = [x for x in self.reactions]
        colors = [self.reactions[x] for x in self.reactions]
        objs += [x for x in self.compounds]
        colors += [self.compounds[x] for x in self.compounds]
                
        return objs,colors
    
    def getBorders(self):
        '''
        Returns a tuple --> objects, color
        '''
        objs = [x for x in self.reactions]
        objs += [x for x in self.compounds]
        
        colors = []
        for x in list(self.reactions.keys())+list(self.compounds.keys()):
            if x in list(self.borders.keys()):
                colors.append(self.borders[x])
            else:
                colors.append(None)
                
        return objs,colors
        
class KeggDetails(object):
    '''
    Class KoDetails
    All the informations returned by Mappers are contained here
    '''
    def __init__(self):
        # Details
        self.ko = None
        self.react = None
        self.comp = None
        self.path = None
        self.rpair = None
        # Links
        self.koreact = None
        self.pathreact = None
        self.pathcomp = None
        self.compreact = None
        self.reactcomp = None
        self.reactrpair = None
        self.rpairreact = None
        # Maps
        self.pathmaps = None
    
    def _purgeDetails(self,det):
        erase = []
        if not det:
            return det
        
        for key, value in list(det.items()):
            if not value:
                erase.append(key)
        
        for key in erase:
            del det[key]
        
        return det
    
    def setDetails(self, ko=None, react=None, comp=None, path=None, rpair=None):
        self.ko = self._purgeDetails(ko)
        self.react = self._purgeDetails(react)
        self.comp = self._purgeDetails(comp)
        self.path = self._purgeDetails(path)
        self.rpair = self._purgeDetails(rpair)
        
    def setLinks(self, koreact=None, pathreact=None, pathcomp=None,
                 compreact=None, reactcomp=None, reactrpair=None,
                 rpairreact=None):
        self.koreact = {}
        if koreact:
            for k,v in list(koreact.items()):
                self.koreact[k] = []
                for i in v:
                    self.koreact[k].append(str(i))

        self.pathreact = {}
        if pathreact:
            for k,v in list(pathreact.items()):
                self.pathreact[k] = []
                for i in v:
                    self.pathreact[k].append(str(i))
                    
        self.pathcomp = {}
        if pathcomp:
            for k,v in list(pathcomp.items()):
                self.pathcomp[k] = []
                for i in v:
                    self.pathcomp[k].append(str(i))
                    
        self.compreact = {}
        if compreact:
            for k,v in list(compreact.items()):
                self.compreact[k] = []
                for i in v:
                    self.compreact[k].append(str(i))
                    
                self.compreact = {}
        
        self.reactcomp = {}
        if reactcomp:
            for k,v in list(reactcomp.items()):
                self.reactcomp[k] = []
                for i in v:
                    self.reactcomp[k].append(str(i))
                    
        self.reactrpair = {}
        if reactrpair:
            for k,v in list(reactrpair.items()):
                self.reactrpair[k] = []
                for i in v:
                    self.reactrpair[k].append(str(i))
                    
        self.rpairreact = {}
        if rpairreact:
            for k,v in list(rpairreact.items()):
                self.rpairreact[k] = []
                for i in v:
                    self.rpairreact[k].append(str(i))
    
    def setMaps(self, maps):
        self.pathmaps = maps
    
    def getKO(self):
        return self.ko
    
    def getReact(self):
        return self.react
    
    def getComp(self):
        return self.comp
    
    def getPath(self):
        return self.path
    
    def getKOLinks(self):
        return self.koreact
    
    def getCompLinks(self):
        return self.comppath
    
    def getPathLinks(self):
        return self.pathreact, self.pathcomp
    
    def getMaps(self):
        return self.pathmaps

class BaseKegg(CommonThread):
    def __init__(self, threads=10, keeptrying=False, queue=queue.Queue()):
        CommonThread.__init__(self,queue)
        
        # Kegg connection
        self.handlers = []
        self._hindex = 0
        self.numThreads = threads
        
        for i in range(self.numThreads):
            obj = KeggAPI(keeptrying)
            self.handlers.append(obj)
            
        self.cleanHandlers()
    
    def cleanHandlers(self):
        for handler in self.handlers:
            handler.clean()
        self._hindex = 0
    
    def getHandler(self):
        if self._hindex >= len(self.handlers):
            self._hindex = 0
        handler = self.handlers[self._hindex] 
        self._hindex += 1
        return handler
        
    def checkConnection(self):
        '''
        Check if there are connection problems
        First check the two IP addresses, then the URL
        '''
        check = [KeggAPI().baseurl, KeggAPI()._apiurl]
        online = False
        for addr in check:
            try:
                isOnline(addr)
                online = True
            except:
                logger.debug('address %s not working'%addr)        
        if not online:
            raise Exception('KEGG seems to be offline')
            
class BaseMapper(BaseKegg):
    def __init__(self, threads=10, avoid=[], keeptrying=False,
                        queue=queue.Queue()):
        BaseKegg.__init__(self, threads=threads, keeptrying=keeptrying,
                                queue=queue)

        # Skip these
        self.avoid = avoid
        
        # Results
        self.reactdet = {}
        self.rpairdet = {}
        self.pathdet = {}
        self.pathreact = {}
        self.pathcomp = {}
        self.pathmap = {}
        self.compdet = {}
        self.reactpath = {}
        self.reactcomp = {}
        self.compreact = {}
        self.rpairreact = {}
        self.reactrpair = {}
        
        # Output
        self.result = None
        
    def getReactDetails(self):
        pieces = [p for p in get_span(list(self.reactdet.keys()), 9)]
        for piece in get_span(pieces, self.numThreads):
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return
            
            self.cleanHandlers()
            self._substatus += len([i for p in piece for i in p])
            if self._substatus > self._maxsubstatus:
                self._substatus = self._maxsubstatus
            self.updateStatus(sub=True)
            
            threads = []
            for ids in piece:
                remove = set()
                for i in ids:
                    if i in self.avoid:
                        remove.add(i)
                for i in remove:
                    ids.remove(i)
                
                if len(ids) == 0:
                    continue
                
                obj = threading.Thread(
                                target = self.getHandler().getTitle,
                                args = (ids,['ENZYME'],))
                obj.start()
                threads.append(obj)
            time.sleep(0.01)
            
            if len(threads) == 0:
                continue
            
            while len(threads) > 0:
                for thread in threads:
                    if not thread.isAlive():
                        threads.remove(thread)
            for handler in self.handlers:
                if handler.failed:
                    logger.error('KEGG API error, aborting')
                    raise IOError('KEGG API error')
                
                if not handler.result:
                    logger.debug('Found an empty handler')
                    continue
                
                for kid, title in list(handler.result.items()):
                    self.reactdet[kid] = title
    
    def getRPairDetails(self):
        pieces = [p for p in get_span(list(self.rpairdet.keys()), 9)]
        for piece in get_span(pieces, self.numThreads):
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return
            
            self.cleanHandlers()
            self._substatus += len([i for p in piece for i in p])
            if self._substatus > self._maxsubstatus:
                self._substatus = self._maxsubstatus
            self.updateStatus(sub=True)
            
            for ids in piece:
                remove = set()
                for i in ids:
                    if i in self.avoid:
                        remove.add(i)
                for i in remove:
                    ids.remove(i)
                    
                if len(ids) == 0:
                    continue
                
                for rid in ids:
                    self.rpairdet[rid] = [rid.split('_')[0],
                                          rid.split('_')[1],
                                          'main']
    
    def getPathDetails(self):
        pieces = [p for p in get_span(list(self.pathdet.keys()), 9)]
        for piece in get_span(pieces, self.numThreads):
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return
            
            self.cleanHandlers()
            self._substatus += len([i for p in piece for i in p])
            if self._substatus > self._maxsubstatus:
                self._substatus = self._maxsubstatus
            self.updateStatus(sub=True)
            
            threads = []
            for ids in piece:
                remove = set()
                for i in ids:
                    if i in self.avoid:
                        remove.add(i)
                for i in remove:
                    ids.remove(i)
                    
                if len(ids) == 0:
                    continue
                
                obj = threading.Thread(
                                target = self.getHandler().getTitle,
                                args = (ids,))
                obj.start()
                threads.append(obj)
            time.sleep(0.01)
            
            if len(threads) == 0:
                continue
            
            while len(threads) > 0:
                for thread in threads:
                    if not thread.isAlive():
                        threads.remove(thread)
            for handler in self.handlers:
                if handler.failed:
                    logger.error('KEGG API error, aborting')
                    raise IOError('KEGG API error')
                
                if not handler.result:
                    logger.debug('Found an empty handler')
                    continue
                
                for kid, title in list(handler.result.items()):
                    self.pathdet[kid] = title
    
    def getMapsDetails(self):
        for piece in get_span(list(self.pathdet.keys()), self.numThreads):
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return
            
            self.cleanHandlers()
            self._substatus += self.numThreads
            if self._substatus > self._maxsubstatus:
                self._substatus = self._maxsubstatus
            self.updateStatus(sub=True)
            
            threads = []
            for path in piece:
                if path in self.avoid:
                    continue
                
                obj = threading.Thread(
                            target = self.handlers[piece.index(path)].getHTMLColoredPathway,
                        args = (path,[],[],))
                obj.start()
                threads.append(obj)
            time.sleep(0.01)
            
            if len(threads) == 0:
                continue
            
            while len(threads) > 0:
                for thread in threads:
                    if not thread.isAlive():
                        threads.remove(thread)
            for handler in self.handlers:
                if handler.failed:
                    logger.error('KEGG API error, aborting')
                    raise IOError('KEGG API error')
                
                if not handler.result:
                    logger.debug('Found an empty handler')
                    continue
                parser = MapParser(handler.result)
                self.pathmap[handler.input] = parser.map
    
    def getPathReactions(self):
        pieces = [p for p in get_span(list(self.pathdet.keys()), 80)]
        for piece in get_span(pieces, self.numThreads):
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return
            
            self.cleanHandlers()
            self._substatus += len([i for p in piece for i in p])
            if self._substatus > self._maxsubstatus:
                self._substatus = self._maxsubstatus
            self.updateStatus(sub=True)
            
            threads = []
            for ids in piece:
                remove = set()
                for i in ids:
                    if i in self.avoid:
                        remove.add(i)
                for i in remove:
                    ids.remove(i)
                    
                if len(ids) == 0:
                    continue
                
                obj = threading.Thread(
                                target = self.getHandler().getReactionsFromPath,
                                args = (ids,))
                obj.start()
                threads.append(obj)
            time.sleep(0.01)
            
            if len(threads) == 0:
                continue
            
            while len(threads) > 0:
                for thread in threads:
                    if not thread.isAlive():
                        threads.remove(thread)
            for handler in self.handlers:
                if handler.failed:
                    logger.error('KEGG API error, aborting')
                    raise IOError('KEGG API error')
                
                if not handler.result:
                    logger.debug('Found an empty handler')
                    continue
                
                for path, reacts in list(handler.result.items()):
                    if path not in self.pathreact:
                        self.pathreact[path] = reacts
                reacts = set([v for vs in list(handler.result.values()) for v in vs])
                for react in reacts:
                    if react not in self.reactdet:
                        self.reactdet[react] = None
                        
    def getPathCompounds(self):
        pieces = [p for p in get_span(list(self.pathdet.keys()), 80)]
        for piece in get_span(pieces, self.numThreads):
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return
            
            self.cleanHandlers()
            self._substatus += len([i for p in piece for i in p])
            if self._substatus > self._maxsubstatus:
                self._substatus = self._maxsubstatus
            self.updateStatus(sub=True)
            
            threads = []
            for ids in piece:
                remove = set()
                for i in ids:
                    if i in self.avoid:
                        remove.add(i)
                for i in remove:
                    ids.remove(i)
                    
                if len(ids) == 0:
                    continue
                
                obj = threading.Thread(
                                target = self.getHandler().getCompoundsFromPath,
                                args = (ids,))
                obj.start()
                threads.append(obj)
            time.sleep(0.01)
            
            if len(threads) == 0:
                continue
            
            while len(threads) > 0:
                for thread in threads:
                    if not thread.isAlive():
                        threads.remove(thread)
            for handler in self.handlers:
                if handler.failed:
                    logger.error('KEGG API error, aborting')
                    raise IOError('KEGG API error')
                
                if not handler.result:
                    logger.debug('Found an empty handler')
                    continue
                
                for path, comps in list(handler.result.items()):
                    if path not in self.pathcomp:
                        self.pathcomp[path] = comps
                comps = set([v for vs in list(handler.result.values()) for v in vs])
                for comp in comps:
                    if comp not in self.compdet:
                        self.compdet[comp] = None
                        
    def getCompDetails(self):
        pieces = [p for p in get_span(list(self.compdet.keys()), 9)]
        for piece in get_span(pieces, self.numThreads):
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return
            
            self.cleanHandlers()
            self._substatus += len([i for p in piece for i in p])
            if self._substatus > self._maxsubstatus:
                self._substatus = self._maxsubstatus
            self.updateStatus(sub=True)
            
            threads = []
            for ids in piece:
                remove = set()
                for i in ids:
                    if i in self.avoid:
                        remove.add(i)
                for i in remove:
                    ids.remove(i)
                    
                if len(ids) == 0:
                    continue
                
                obj = threading.Thread(
                                target = self.getHandler().getTitle,
                                args = (ids,))
                obj.start()
                threads.append(obj)
            time.sleep(0.01)
            
            if len(threads) == 0:
                continue
            
            while len(threads) > 0:
                for thread in threads:
                    if not thread.isAlive():
                        threads.remove(thread)
            for handler in self.handlers:
                if handler.failed:
                    logger.error('KEGG API error, aborting')
                    raise IOError('KEGG API error')
                
                if not handler.result:
                    logger.debug('Found an empty handler')
                    continue
                
                for kid, title in list(handler.result.items()):
                    self.compdet[kid] = title
    
    def getPathways(self):
        pieces = [p for p in get_span(list(self.reactdet.keys()), 80)]
        for piece in get_span(pieces, self.numThreads):
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return
            
            self.cleanHandlers()
            self._substatus += len([i for p in piece for i in p])
            if self._substatus > self._maxsubstatus:
                self._substatus = self._maxsubstatus
            self.updateStatus(sub=True)
            
            threads = []
            for ids in piece:
                remove = set()
                for i in ids:
                    if i in self.avoid:
                        remove.add(i)
                for i in remove:
                    ids.remove(i)
                    
                if len(ids) == 0:
                    continue
                
                obj = threading.Thread(
                                target = self.getHandler().getPathways,
                                args = (ids,))
                obj.start()
                threads.append(obj)
            time.sleep(0.01)
            
            if len(threads) == 0:
                continue
            
            while len(threads) > 0:
                for thread in threads:
                    if not thread.isAlive():
                        threads.remove(thread)
            for handler in self.handlers:
                if handler.failed:
                    logger.error('KEGG API error, aborting')
                    raise IOError('KEGG API error')
                
                if not handler.result:
                    logger.debug('Found an empty handler')
                    continue
                
                for react, paths in list(handler.result.items()):
                    if react not in self.reactpath:
                        self.reactpath[react] = []
                    for path in paths:
                        if path.startswith('path:map'):continue
                        self.reactpath[react].append(path)
                paths = set([v for vs in list(handler.result.values()) for v in vs])
                for path in paths:
                    if path not in self.pathdet and not path.startswith('path:map'):
                        self.pathdet[path] = None
                        
    def getReactCompounds(self):
        pieces = [p for p in get_span(list(self.reactdet.keys()), 80)]
        for piece in get_span(pieces, self.numThreads):
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return
            
            self.cleanHandlers()
            self._substatus += len([i for p in piece for i in p])
            if self._substatus > self._maxsubstatus:
                self._substatus = self._maxsubstatus
            self.updateStatus(sub=True)
            
            threads = []
            for ids in piece:
                remove = set()
                for i in ids:
                    if i in self.avoid:
                        remove.add(i)
                for i in remove:
                    ids.remove(i)
                    
                if len(ids) == 0:
                    continue
                
                obj = threading.Thread(
                                target = self.getHandler().getCompoundsFromReaction,
                                args = (ids,))
                obj.start()
                threads.append(obj)
            time.sleep(0.01)
            
            if len(threads) == 0:
                continue
            
            while len(threads) > 0:
                for thread in threads:
                    if not thread.isAlive():
                        threads.remove(thread)
            for handler in self.handlers:
                if handler.failed:
                    logger.error('KEGG API error, aborting')
                    raise IOError('KEGG API error')
                
                if not handler.result:
                    logger.debug('Found an empty handler')
                    continue
                
                for react, comps in list(handler.result.items()):
                    if react not in self.reactcomp:
                        self.reactcomp[react] = comps
                comps = set([v for vs in list(handler.result.values()) for v in vs])
                for comp in comps:
                    if comp not in self.compdet:
                        self.compdet[comp] = None
                        
    def getCompoundReacts(self):
        pieces = [p for p in get_span(list(self.compdet.keys()), 80)]
        for piece in get_span(pieces, self.numThreads):
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return
            
            self.cleanHandlers()
            self._substatus += len([i for p in piece for i in p])
            if self._substatus > self._maxsubstatus:
                self._substatus = self._maxsubstatus
            self.updateStatus(sub=True)
            
            threads = []
            for ids in piece:
                remove = set()
                for i in ids:
                    if i in self.avoid:
                        remove.add(i)
                for i in remove:
                    ids.remove(i)
                    
                if len(ids) == 0:
                    continue
                
                obj = threading.Thread(
                                target = self.getHandler().getReactionsByComp,
                                args = (ids,))
                obj.start()
                threads.append(obj)
            time.sleep(0.01)
            
            if len(threads) == 0:
                continue
            
            while len(threads) > 0:
                for thread in threads:
                    if not thread.isAlive():
                        threads.remove(thread)
            for handler in self.handlers:
                if handler.failed:
                    logger.error('KEGG API error, aborting')
                    raise IOError('KEGG API error')
                
                if not handler.result:
                    logger.debug('Found an empty handler')
                    continue
                
                for comp, reacts in list(handler.result.items()):
                    if comp not in self.compreact:
                        self.compreact[comp] = reacts
                reacts = set([v for vs in list(handler.result.values()) for v in vs])
                for react in reacts:
                    if react not in self.reactdet:
                        self.reactdet[react] = None
    
    def getReactRPairs(self):
        pieces = [p for p in get_span(list(self.reactdet.keys()), 80)]
        for piece in get_span(pieces, self.numThreads):
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return
            
            self.cleanHandlers()
            self._substatus += len([i for p in piece for i in p])
            if self._substatus > self._maxsubstatus:
                self._substatus = self._maxsubstatus
            self.updateStatus(sub=True)
            
            threads = []
            for ids in piece:
                remove = set()
                for i in ids:
                    if i in self.avoid:
                        remove.add(i)
                for i in remove:
                    ids.remove(i)
                    
                if len(ids) == 0:
                    continue
                
                obj = threading.Thread(
                                target = self.getHandler().getRPairsFromReaction,
                                args = (ids,))
                obj.start()
                threads.append(obj)
            time.sleep(0.01)
            
            if len(threads) == 0:
                continue
            
            while len(threads) > 0:
                for thread in threads:
                    if not thread.isAlive():
                        threads.remove(thread)
            for handler in self.handlers:
                if handler.failed:
                    logger.error('KEGG API error, aborting')
                    raise IOError('KEGG API error')
                
                if not handler.result:
                    logger.debug('Found an empty handler')
                    continue
                
                for react, rpairs in list(handler.result.items()):
                    if react not in self.reactrpair:
                        self.reactrpair[react] = rpairs
                rpairs = set([v for vs in list(handler.result.values()) for v in vs])
                for rpair in rpairs:
                    if rpair not in self.rpairdet:
                        self.rpairdet[rpair] = None

class KoMapper(BaseMapper):
    '''
    From a list of KO IDs returns various details in an object
    KO --> title and details
    KO --> reactions (and titles)
    reactions --> pathways (and titles)
    pathways --> reactions, compounds (with titles)
    maps --> for each pathway, the html maps (!!!)
    '''
    
    _statusDesc = {0:'Not started',
               1:'Checking connectivity',
               2:'Fetching reactions',
               3:'Fetching rpairs',
               4:'Fetching pathways',
               5:'Fetching pathways content',
               6:'Fetching reactions - compounds links',
               7:'Fetching details on KEGG entries',
               8:'Crafting results'}
    
    _substatuses = [2,3,4,5,6,7]
    
    def __init__(self, ko_list, threads=40, avoid=[], keeptrying=False,
                        queue=queue.Queue()):
        BaseMapper.__init__(self, threads=threads, avoid=avoid,
                            keeptrying=keeptrying, queue=queue)
        # Kegg
        self.ko = ko_list
        
        # Results
        self.kodet = {}
        self.koreact = {}
    
    def getKOdet(self):
        pieces = [p for p in get_span(self.ko, 9)]
        for piece in get_span(pieces, self.numThreads):
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return
            
            self.cleanHandlers()
            self._substatus += len([i for p in piece for i in p])
            if self._substatus > self._maxsubstatus:
                self._substatus = self._maxsubstatus
            self.updateStatus(sub=True)
            
            threads = []
            for ids in piece:
                remove = set()
                for i in ids:
                    if i in self.avoid:
                        remove.add(i)
                for i in remove:
                    ids.remove(i)
                    
                if len(ids) == 0:
                    continue
                
                obj = threading.Thread(
                                target = self.getHandler().getTitle,
                                args = (ids,))
                obj.start()
                threads.append(obj)
            time.sleep(0.01)
            
            if len(threads) == 0:
                continue
            
            while len(threads) > 0:
                for thread in threads:
                    if not thread.isAlive():
                        threads.remove(thread)
            for handler in self.handlers:
                if handler.failed:
                    logger.error('KEGG API error, aborting')
                    raise IOError('KEGG API error')
                
                if not handler.result:
                    logger.debug('Found an empty handler')
                    continue
                
                for kid, title in list(handler.result.items()):
                    self.kodet[kid] = title
                
    def getReactions(self):
        pieces = [p for p in get_span(self.ko, 80)]
        for piece in get_span(pieces, self.numThreads):
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return
            
            self.cleanHandlers()
            self._substatus += len([i for p in piece for i in p])
            if self._substatus > self._maxsubstatus:
                self._substatus = self._maxsubstatus
            self.updateStatus(sub=True)
            
            threads = []
            for ids in piece:
                remove = set()
                for i in ids:
                    if i in self.avoid:
                        remove.add(i)
                for i in remove:
                    ids.remove(i)
                    
                if len(ids) == 0:
                    continue
                
                obj = threading.Thread(
                                target = self.getHandler().getReactions,
                                args = (ids,))
                obj.start()
                threads.append(obj)
            time.sleep(0.01)
            
            if len(threads) == 0:
                continue
            
            while len(threads) > 0:
                for thread in threads:
                    if not thread.isAlive():
                        threads.remove(thread)
            for handler in self.handlers:
                if handler.failed:
                    logger.error('KEGG API error, aborting')
                    raise IOError('KEGG API error')
                
                if not handler.result:
                    logger.debug('Found an empty handler')
                    continue
                
                for ko, reacts in list(handler.result.items()):
                    if ko not in self.koreact:
                        self.koreact[ko] = reacts
                reacts = set([v for vs in list(handler.result.values()) for v in vs])
                for react in reacts:
                    if react not in self.reactdet:
                        self.reactdet[react] = None
    
    def run(self):
        self.updateStatus()
        try:
            self.checkConnection()
        except Exception as e:
            self.sendFailure(str(e))
            return
    
        # Reactions
        self._maxsubstatus = len(self.ko)
        self.updateStatus()
        try:
            self.getReactions()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Related rpairs
        self._maxsubstatus = len(self.reactdet)
        self.updateStatus()
        logger.warning('Using RCLASS attribute of KEGG reactions, as the RPAIR database has now been discontinued')
        try:
            self.getReactRPairs()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Related pathways...
        self._maxsubstatus = len(self.reactdet)
        self.updateStatus()
        try:
            self.getPathways()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Pathways contents...
        # 1. Reactions
        self._maxsubstatus = len(self.pathdet)
        self.updateStatus()
        try:
            self.getPathReactions()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # 2. Compounds
        self._maxsubstatus = len(self.pathdet)
        try:
            self.getPathCompounds()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Compounds for each reaction
        self._maxsubstatus = len(self.reactdet)
        self.updateStatus()
        try:
            self.getReactCompounds()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Reactions for each compound
        self._maxsubstatus = len(self.compdet)
        try:
            self.getCompoundReacts()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # KO details
        self._maxsubstatus = len(self.ko)
        self.updateStatus()
        try:
            self.getKOdet()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Pathway details
        self._maxsubstatus = len(self.pathdet)
        try:
            self.getPathDetails()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Pathway HTML maps (!!!)
        self._maxsubstatus = len(self.pathdet)
        try:
            self.getMapsDetails()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Reaction details
        self._maxsubstatus = len(self.reactdet)
        try:
            self.getReactDetails()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Compound details
        self._maxsubstatus = len(self.compdet)
        try:
            self.getCompDetails()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # RPair details
        self._maxsubstatus = len(self.rpairdet)
        logger.warning('Using RCLASS attribute of KEGG reactions, as the RPAIR database has now been discontinued')
        try:
            self.getRPairDetails()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Prepare the output object
        self.updateStatus()
        self.result = KeggDetails()
        self.result.setDetails(self.kodet, self.reactdet,
                               self.compdet, self.pathdet, self.rpairdet)
        self.result.setLinks(koreact=self.koreact, pathreact=self.pathreact, 
                             pathcomp=self.pathcomp, reactcomp=self.reactcomp,
                             compreact=self.compreact,
                             rpairreact=self.rpairreact,
                             reactrpair=self.reactrpair)
        self.result.setMaps(self.pathmap)

class CompMapper(BaseMapper):
    '''
    From a list of CO IDs returns various details in an object
    CO --> title and details
    CO --> pathways (and titles)
    pathways --> reactions, compounds (with titles)
    maps --> for each pathway, the html maps (!!!)
    '''
    
    _statusDesc = {0:'Not started',
               1:'Checking connectivity',
               2:'Fetching reactions',
               3:'Fetching rpairs',
               4:'Fetching pathways',
               5:'Fetching pathways content',
               6:'Fetching reactions - compounds links',
               7:'Fetching details on KEGG entries',
               8:'Crafting results'}
    
    _substatuses = [2,3,4,5,6,7]
    
    def __init__(self, co_list, threads=40, avoid=[], keeptrying=False,
                        queue=queue.Queue()):
        BaseMapper.__init__(self, threads=threads, avoid=avoid,
                            keeptrying=keeptrying, queue=queue)
        # Kegg
        self.co = co_list
        
        # Results
        self.comppath = {}
    
    def run(self):
        self.updateStatus()
        try:
            self.checkConnection()
        except Exception as e:
            self.sendFailure(str(e))
            return
    
        # Reactions
        for co_id in self.co:
            self.compdet[co_id] = None
        self._maxsubstatus = len(self.compdet)
        self.updateStatus()
        try:
            self.getCompoundReacts()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Related rpairs
        self._maxsubstatus = len(self.reactdet)
        self.updateStatus()
        logger.warning('Using RCLASS attribute of KEGG reactions, as the RPAIR database has now been discontinued')
        try:
            self.getReactRPairs()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Related pathways...
        self._maxsubstatus = len(self.reactdet)
        self.updateStatus()
        try:
            self.getPathways()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Pathways contents...
        # 1. Reactions
        self._maxsubstatus = len(self.pathdet)
        self.updateStatus()
        try:
            self.getPathReactions()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # 2. Compounds
        self._maxsubstatus = len(self.pathdet)
        try:
            self.getPathCompounds()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Compunds for each reaction
        self._maxsubstatus = len(self.reactdet)
        self.updateStatus()
        try:
            self.getReactCompounds()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Reactions for each compund
        self._maxsubstatus = len(self.compdet)
        try:
            self.getCompoundReacts()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Pathway details
        self._maxsubstatus = len(self.pathdet)
        self.updateStatus()
        try:
            self.getPathDetails()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Pathway HTML maps (!!!)
        self._maxsubstatus = len(self.pathdet)
        try:
            self.getMapsDetails()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Reaction details
        self._maxsubstatus = len(self.reactdet)
        try:
            self.getReactDetails()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Compound details
        self._maxsubstatus = len(self.compdet)
        try:
            self.getCompDetails()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # RPair details
        self._maxsubstatus = len(self.rpairdet)
        logger.warning('Using RCLASS attribute of KEGG reactions, as the RPAIR database has now been discontinued')
        try:
            self.getRPairDetails()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Prepare the output object
        self.updateStatus()
        self.result = KeggDetails()
        self.result.setDetails(react=self.reactdet,
                               comp=self.compdet, path=self.pathdet,
                               rpair=self.rpairdet)
        self.result.setLinks(pathreact=self.pathreact, 
                             pathcomp=self.pathcomp, compreact=self.compreact,
                             reactcomp=self.reactcomp,
                             reactrpair=self.reactrpair,
                             rpairreact=self.rpairreact)
        self.result.setMaps(self.pathmap)

class MapsFetcher(BaseKegg):
    '''
    Class MapsFetcher
    Download colored Kegg maps (png or URLs)
    Input: color_objs (KeggColor list), picture, htmls, urls, prefix, legend (file)
    Output: tuple(list of png filenames, list of HTML files, list of URLs)
    '''
    
    _statusDesc = {0:'Not started',
               1:'Checking connectivity',
               2:'Making room',
               3:'Fetching maps (pictures)',
               4:'Generating interactive web pages'}
    
    _substatuses = [3]
    
    def __init__(self, color_objs, pictures=True, html=True, prefix='', 
                 legend=None, threads=40, keeptrying=False,
                 queue=queue.Queue()):
        BaseKegg.__init__(self, threads=threads, keeptrying=keeptrying,
                          queue=queue)
        
        self.colors = color_objs
        self.pictures = bool(pictures)
        self.web = bool(html)
        self.legend = legend 
        
        self._keggroom = None
        self._prefix = prefix 
        
        # Outputs
        self.pics = []
        self.webpages = []
        self.pages = []
        self.result = (self.pics, self.webpages, self.pages)
    
    def makeRoom(self,location=''):
        '''
        Creates a tmp directory in the desired location
        '''
              
        # KEGG database path
        try:
            path = os.path.abspath(location)
            path = os.path.join(path, 'tmp')
            try:os.mkdir(path)
            except:pass
            path = os.path.join(path, 'keggmaps')
            try:os.mkdir(path)
            except:pass
            path = os.path.join(path, self._prefix)
            self._keggroom = path
            os.mkdir(path)
        except:
            logger.debug('Temporary directory creation failed! %s'
                          %path)
    
    def copyLegend(self):
        '''Copy the legend in the target directory'''
        if self.legend and os.path.exists(self.legend):
            legend = os.path.join(self._keggroom, 'legend.png')
            shutil.copyfile(self.legend, legend)
            
            return legend
        return None
    
    def getMaps(self):
        legend = self.copyLegend()
        
        for piece in get_span(self.colors, self.numThreads):
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return
            
            self.cleanHandlers()
            self._substatus += self.numThreads
            if self._substatus > self._maxsubstatus:
                self._substatus = self._maxsubstatus
            self.updateStatus(sub=True)
            
            threads = []
            for kmap in piece:
                path = kmap.path
                
                # Skip the general maps
                if path in avoidedPaths:
                    logger.debug('Skipping general pathway %s'%path)
                    continue
                #
                
                objs,colors = kmap.getAll()
                dummy,borders = kmap.getBorders()
                
                obj = threading.Thread(
                        target = self.handlers[piece.index(kmap)].getHTMLColoredPathway,
                        args = (path,objs,colors,borders,))
                obj.start()
                threads.append(obj)
            time.sleep(0.01)
            
            if len(threads) == 0:
                continue
            
            while len(threads) > 0:
                for thread in threads:
                    if not thread.isAlive():
                        threads.remove(thread)
            for handler in self.handlers:
                if handler.failed:
                    logger.error('KEGG API error, aborting')
                    raise IOError('KEGG API error')
                
                if not handler.result:
                    logger.debug('Found an empty handler')
                    continue
                
                fname = os.path.join(self._keggroom,handler.input)
                fname = fname+'.png'
                
                # Fetch the map picture
                # Hoping it won't change much in the future
                if isinstance(handler.result, bytes):
                    handler.result = handler.result.decode('utf-8')
                for line in handler.result.split('\n'):
                    if ('<img' in line
                        and 'pathwayimage' in line
                        and 'usemap="#mapdata"' in line):
                        urlimage = 'http://www.kegg.jp/' + line.split('src="')[1].split('"')[0]

                        sock=urlopen(urlimage, timeout=30)
                        pic = sock.read()
                        #if isinstance(pic, bytes):
                        #    pic = pic.decode('utf-8')
                        sock.close()
                        
                        fOut = open(fname,'wb')
                        fOut.write(pic)
                        fOut.close()
                        self.pics.append(fname)
                #
    
    def getWebPages(self):
        # TODO: nicer web pages
        legend = self.copyLegend()
        legend = os.path.split(legend)[-1]
        
        if legend:
            fname = os.path.join(self._keggroom,'legend.html')
            
            fOut = open(fname,'w')
            fOut.write('<html>\n<head></head>\n<body>\n')
            fOut.write('''<div align="center">
                        <img src="./%s" />
                        </div>\n'''%
                       (legend))
            fOut.write('</body>\n</html>')
            fOut.close()
        
        for path in self.colors:
            myindex = self.colors.index(path)
            
            logger.debug('Writing interactive web page for %s'%path.path)
            
            fname = os.path.join(self._keggroom,path.path)
            fname = fname+'.html'
            
            fOut = open(fname,'w')
            fOut.write('<html>\n%s\n<body>\n'%kheader)
            
            # Navigation
            fOut.write('''<h2 align="center">\n''')
            fOut.write('''<a href="./%s.html">&laquo;</a>%s'''%(
                       self.colors[myindex-1].path, path.path))
            try:
                next = self.colors[myindex+1]
            except:
                next = self.colors[0]
            fOut.write('''<a href="./%s.html">&raquo;</a>\n</h2>\n'''%
                       (next.path))
            
            if legend:
                fOut.write('''<h3 align="center">
                            <a href="./legend.html">Color scheme</a>
                            </h3>\n''')
            
            fOut.write('''<div align="center">
                        <img src="./%s" usemap="#mapdata" border="0" />
                        </div>\n'''%
                       (path.path+'.png'))
            
            html = path.htmlmap.split('\n')
            newhtml = []
            for line in html:
                line = line.replace('href="/dbget-bin/www_bget?',
               'target="_blank" href="http://www.genome.jp/dbget-bin/www_bget?')
                
                if '/kegg-bin/show_pathway?' in line:
                    s = line.split('/kegg-bin/show_pathway?')
                    s1 = s[1].split('"')
                    s1[0] += '.html'
                    line1 = '"'.join(s1)
                    line = './'.join([s[0]] + [line1])
                
                newhtml.append(line)
                
            fOut.write('%s\n'%'\n'.join(newhtml))
            fOut.write('<div id="poplay" class="poplay" />\n</body>\n</html>')
            fOut.close()
            
            self.webpages.append(fname)
    
    def run(self):
        self.updateStatus()
        try:
            self.checkConnection()
        except Exception as e:
            self.sendFailure(str(e))
            return
    
        self.updateStatus()
        self.makeRoom()
        
        if self.killed:
            return
        
        # ':' bugfix
        # the ':' char causes various problems in windows folders
        for path in self.colors:
            if ':' in path.path:
                path.path = path.path.split(':')[1]
        
        if self.pictures:
            self._maxsubstatus = len(self.colors)
            self.updateStatus()
            try:
                self.getMaps()
            except Exception as e:
                self.sendFailure(e)
                return
            self.cleanHandlers()
            self.resetSubStatus()
        else:
            self.updateStatus(send=False)
            
        if self.killed:
            return
        
        if self.web:
            self.updateStatus()
            try:
                self.getWebPages()
            except Exception as e:
                self.sendFailure(e)
                return
        else:
            self.updateStatus(send=False)

class KeggNet(BaseMapper):
    '''
    Fetch as much details as possible from the KEGG database,
    starting from the pathways list
    '''
    
    _statusDesc = {0:'Not started',
               1:'Checking connectivity',
               2:'Fetching pathways',
               3:'Fetching compounds',
               4:'Fetching compounds - reactions links',
               5:'Fetching reactions',
               6:'Fetching rpairs',
               7:'Fetching reactions - compounds links',
               8:'Fetching details on KEGG entries',
               9:'Crafting results'}
    
    _substatuses = [3,4,5,6,7,8]
    
    def __init__(self, threads=40, avoid=[], keeptrying=False,
                        queue=queue.Queue()):
        BaseMapper.__init__(self, threads=threads, avoid=avoid,
                                    keeptrying=keeptrying, queue=queue)
        
    def getAllPathways(self):
        '''
        Get all the available pathway IDs
        '''
        kegg = KeggAPI()
        
        kegg.getIDListFromDB('pathway')
        
        for p in kegg.result:
            self.pathdet[p] = None
    
    def run(self):
        self.updateStatus()
        try:
            self.checkConnection()
        except Exception as e:
            self.sendFailure(str(e))
            return
    
        # Get pathways
        self.updateStatus()
        try:
            self.getAllPathways()
        except Exception as e:
            self.sendFailure(str(e))
            return
        
        if self.killed:
            return
        
        # Compounds
        self._maxsubstatus = len(self.pathdet)
        self.updateStatus()
        try:
            self.getPathCompounds()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Reactions for each compound
        self._maxsubstatus = len(self.compdet)
        self.updateStatus()
        try:
            self.getCompoundReacts()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Reactions
        self._maxsubstatus = len(self.pathdet)
        self.updateStatus()
        try:
            self.getPathReactions()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Related rpairs
        self._maxsubstatus = len(self.reactdet)
        self.updateStatus()
        logger.warning('Using RCLASS attribute of KEGG reactions, as the RPAIR database has now been discontinued')
        try:
            self.getReactRPairs()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Compounds for each reaction
        self._maxsubstatus = len(self.reactdet)
        self.updateStatus()
        try:
            self.getReactCompounds()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Details
        # Pathway details
        self._maxsubstatus = len(self.pathdet)
        self.updateStatus()
        try:
            self.getPathDetails()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Pathway HTML maps (!!!)
        self._maxsubstatus = len(self.pathdet)
        try:
            self.getMapsDetails()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Reaction details
        self._maxsubstatus = len(self.reactdet)
        try:
            self.getReactDetails()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Compound details
        self._maxsubstatus = len(self.compdet)
        try:
            self.getCompDetails()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # RPair details
        self._maxsubstatus = len(self.rpairdet)
        logger.warning('Using RCLASS attribute of KEGG reactions, as the RPAIR database has now been discontinued')
        try:
            self.getRPairDetails()
        except Exception as e:
            self.sendFailure(str(e))
            return
        self.cleanHandlers()
        self.resetSubStatus()
        
        if self.killed:
            return
        
        # Prepare the output object
        self.updateStatus()
        self.result = KeggDetails()
        self.result.setDetails(None, self.reactdet,
                               self.compdet, self.pathdet, self.rpairdet)
        self.result.setLinks(pathreact=self.pathreact, 
                             pathcomp=self.pathcomp, reactcomp=self.reactcomp,
                             compreact=self.compreact,
                             rpairreact=self.rpairreact,
                             reactrpair=self.reactrpair)
        self.result.setMaps(self.pathmap)



#!/usr/bin/env python
"""
fitting

Phenome library

biolog data fitting functions
"""
from ductape.common.utils import compress, smooth
try:
    from scipy.optimize.minpack import curve_fit
except ImportError:
    # Old version of SciPy, manual addition of curve_fit
    
    # This three functions are borrowed from the scipy package, licensed under
    # the 3-clauses BSD licence
    def _general_function(params, xdata, ydata, function):
        return function(xdata, *params) - ydata
    
    def _weighted_general_function(params, xdata, ydata, function, weights):
        return weights * (function(xdata, *params) - ydata)
    
    def curve_fit(f, xdata, ydata, p0=None, sigma=None, **kw):
        from numpy import isscalar, asarray, array, inf
        from scipy.optimize import leastsq
        
        if p0 is None:
            # determine number of parameters by inspecting the function
            import inspect
            args, varargs, varkw, defaults = inspect.getargspec(f)
            if len(args) < 2:
                msg = "Unable to determine number of fit parameters."
                raise ValueError(msg)
            if 'self' in args:
                p0 = [1.0] * (len(args)-2)
            else:
                p0 = [1.0] * (len(args)-1)
    
        if isscalar(p0):
            p0 = array([p0])
    
        args = (xdata, ydata, f)
        if sigma is None:
            func = _general_function
        else:
            func = _weighted_general_function
            args += (1.0/asarray(sigma),)
    
        # Remove full_output from kw, otherwise we're passing it in twice.
        return_full = kw.pop('full_output', False)
        res = leastsq(func, p0, args=args, full_output=1, **kw)
        (popt, pcov, infodict, errmsg, ier) = res
    
        if ier not in [1, 2, 3, 4]:
            msg = "Optimal parameters not found: " + errmsg
            raise RuntimeError(msg)
    
        if (len(ydata) > len(p0)) and pcov is not None:
            s_sq = (func(popt, *args)**2).sum()/(len(ydata)-len(p0))
            pcov = pcov * s_sq
        else:
            pcov = inf
    
        if return_full:
            return popt, pcov, infodict, errmsg, ier
        else:
            return popt, pcov
    # End of borrowed scipy fix
        
import numpy as np
import logging
# No country for warnings
import scipy as sp
sp.seterr(all='ignore')
#

__author__ = "Marco Galardini"

logger = logging.getLogger('ductape.fitting')

def logistic(x, A, u, d, v, y0):
    '''
    Logistic growth model
    Taken from: "Modeling of the bacterial growth curve."
                (Zwietering et al., 1990)
                PMID: 16348228
    '''
    y = (A / (1 + np.exp( ( ((4 * u)/A) * (d - x) ) + 2 ))) + y0
    return y

def gompertz(x, A, u, d, v, y0):
    '''
    Gompertz growth model
    Taken from: "Modeling of the bacterial growth curve."
                (Zwietering et al., 1990)
                PMID: 16348228
    '''
    y = (A * np.exp( -np.exp( (((u * np.e)/A) * (d - x)) + 1 ) ) ) + y0
    return y

def richards(x, A, u, d, v, y0):
    '''
    Richards growth model
    (equivalent to Stannard)
    Taken from: "Modeling of the bacterial growth curve."
                (Zwietering et al., 1990)
                PMID: 16348228
    '''
    y = (A * pow(1 + (v + (np.exp(1 + v) * np.exp( (u/A) * (1 + v) * (1 + (1/v)) * (d - x) ) ) ),-(1/v))) + y0
    return y

def getFlex(x, y):
    '''
    Given two axes (with the same length!) returns a guess of the flex point
    '''
    if len(x) != len(y):
        logger.debug('Axes have different sizes (x: %d, y: %d)'%(len(x),len(y)))
        return 0
    
    diffs = []
    indexes = list(range(len(x)))
    
    for i in indexes:
        if i+1 not in indexes:
            continue
        diffs.append(y[i+1] - y[i])
    diffs = np.array( diffs )
    
    flex = x[-1]
    for i in indexes:
        if i+1 not in indexes:
            continue
        if (y[i+1] - y[i]) > (diffs.mean() + (diffs.std())):
            flex = x[i]
            break
    
    return flex

def getPlateau(x, y):
    '''
    Given two axes (with the same length!) returns a guess of the plateau point
    '''
    if len(x) != len(y):
        logger.debug('Axes have different sizes (x: %d, y: %d)'%(len(x),len(y)))
        return 0
    
    ymax = y.max()
    
    diffs = []
    indexes = list(range(len(y)))
    
    for i in indexes:
        if i+1 not in indexes:
            continue
        diffs.append(y[i+1] - y[i])
    diffs = np.array( diffs )
    
    ymax = y[-1]
    for i in indexes:
        if y[i] > (ymax - diffs.std()) and y[i] < (ymax + diffs.std()):
            ymax = y[i]
            break
    
    return ymax
    

def rect(x, a, y0):
    '''
    yep, that's a rect!
    '''
    y = (a * x) + y0
    return y

def fitData(xdata, ydata):
    '''
    Fits the provided data to the first working function
    (first Gompertz, then Logistic, then Richards)

    Returns a tuple with plateau, slope, lag, y0 and model used
    If no fitting was possible all values are None

    Please note that the plateau may be reached outside the final time point
    '''
    retries = 2
    while retries > 0:
        params = [None, None, None, None, None]
        model = ''
        # Initial guesses for the output parameters
        p0 = [getPlateau(xdata, ydata), 4.0, getFlex(xdata, ydata), 0.1, 0]
        if retries == 1:
            p0[2] = 0
        try:
            params, pcov = curve_fit(gompertz, xdata, ydata, p0 = p0)
            model = 'gompertz'
            break
        except:
            #logger.debug('Gompertz fit failed')
            try:
                params, pcov = curve_fit(logistic, xdata, ydata, p0 = p0)
                model = 'logistic'
                break
            except:
                #logger.debug('Logistic fit failed')
                try:
                    params, pcov = curve_fit(richards, xdata, ydata, p0 = p0)
                    model = 'richards'
                    break
                except:
                    #logger.debug('Richards fit failed')
                    retries -= 1
                    #logger.debug('%d retries left'%retries)
                    # Compress again the data
                    ydata = np.array(compress(ydata, span=2))
                    if len(ydata) <= 11:
                        window_len = len(ydata)
                    else:
                        window_len = 11
                    ydata = np.array(smooth(ydata, window_len = window_len, 
                              window = 'blackman'))
                    xdata = np.array(compress(xdata, span=2))
                    #
                    params = [None, None, None, None, None]
    
    return params, model
